"""
client_socket.py:

Socket used to attach to the TCP server as a client and read/write data.
"""

import select
import socket
import threading

from fprime.constants import DATA_ENCODING
from fprime_gds.common.handlers import DataHandler

# Constants for public use
GUI_TAG = "GUI"
FSW_TAG = "FSW"


class ThreadedTCPSocketClient(DataHandler):
    """
    Threaded TCP client that connects to the socket server that serves packets from the flight
    software
    """

    def __init__(self, sock=None, dest=FSW_TAG):
        """
        Threaded client socket constructor

        Keyword Arguments:
                sock {Socket} -- A socket for the client to use. Created own if
                                 None (default: {None})
        """

        if sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock

        # NOTE can't do this b/c EINPROGRESS: self.sock.setblocking(0)
        self.dest = dest
        self.__distributors = []
        self.__select_timeout = 1
        self.__data_recv_thread = threading.Thread(target=self.recv)
        self.stop_event = threading.Event()

    @staticmethod
    def get_data_bytes(string_data):
        """
        Convert the data bytes from string to bytes

        :param string_data: data in string format
        :return: data in bytes format
        """
        return string_data.encode(DATA_ENCODING)

    @staticmethod
    def get_data_string(bytes_data):
        """
        Convert the data bytes from string to bytes

        :param bytes_data: data in bytes format
        :return: data in string format
        """
        return bytes_data.decode(DATA_ENCODING)

    def register_distributor(self, distributor):
        """Registers a fprime.gds.distributor object with this socket

        Arguments:
                fprime.gds.distributor {Distributor} -- Distributor must implement data_callback
        """

        self.__distributors.append(distributor)

    def register_to_server(self, register_as):
        """
        Registers the caller to the server as type register_as

        This function assumes the socket connects to an fprime TCP server

        Args:
            register_as (string): How to identify this process to the TCP server
                                  Can be either "FSW" or "GUI"
        """
        data = "Register %s\n" % register_as

        self.sock.send(self.get_data_bytes(data))

    def connect(self, host, port):
        """Connect to host at given port and start the threaded recv method.

        Arguments:
                host {string} -- IP of the host server
                port {int} -- Port of the host server
        """
        try:
            self.sock.connect((host, port))
            self.__data_recv_thread.start()
        except OSError:
            print("There was a problem connecting to the TCP Server")
            exit(-1)

    def disconnect(self):
        """Disconnect the socket client from the server and stop the internal thread."""
        self.stop_event.set()
        self.__data_recv_thread.join()
        self.sock.close()

    def data_callback(self, data, sender=None):
        """
        Handles incoming data by sending it to a socket.

        :param data: data to send to the client socket
        :param sender: sender source of the data
        """
        self.send(data, self.dest)

    def send(self, data, dest):
        """
        Send data to the server

        All necessary headers are added in this function.

        Arguments:
            data {binary} -- The data to send (What you want the destination
                             to receive)
            dest {String} -- Where to send the data to. Either "FSW" or "GUI"
        """
        self.sock.send(b"A5A5 %s %s" % (self.get_data_bytes(dest), data))

    def recv(self):
        """
        Method run constantly by the enclosing thread. Looks for data from the server.
        """
        while not self.stop_event.is_set():
            ready = select.select([self.sock], [], [], self.__select_timeout)
            if ready[0]:
                chunk = self.sock.recv(1024)
                for d in self.__distributors:
                    d.on_recv(chunk)



"""
Tests format_string util

@Created on March 18, 2021
@janamian
"""

import unittest
from fprime_gds.common.utils.string_util import format_string


class TestFormatString(unittest.TestCase):

    def test_format_with_no_issue(self):
        template = 'Opcode 0x%04X dispatched to port %d and value %f'
        values = (181, 8, 1.234)
        expected = 'Opcode 0x00B5 dispatched to port 8 and value 1.234'
        actual = format_string(template, values)
        self.assertEqual(expected, actual)

    def test_format_value_with_string_input_as_other_types(self):
        template = 'Opcode 0x%04X dispatched to port %d and value %f'
        values = (181, "8", "1.234")
        expected = 'Opcode 0x00B5 dispatched to port 8 and value 1.234'
        actual = format_string(template, values)
        self.assertEqual(expected, actual)

    def test_format_with_format_spec(self):
        template = 'Opcode 0x%04X dispatched to port %04d and value %0.02f'
        values = (181, 8, 1.234)
        expected = 'Opcode 0x00B5 dispatched to port 0008 and value 1.2'
        actual = format_string(template, values)
        self.assertEqual(expected, actual)

    def test_format_bad_case(self):
        template = 'Opcode 0x%04X dispatched with value %f'
        values = ("181", "0.123")
        with self.assertRaises(ValueError):
            self.assertEqual(format_string(template, values))


if __name__ == '__main__':
    unittest.main()



"""
histories.py:

Module used to handle the wrangling of histories for the standard pipeline. This allows the standard pipeline, and other
to compose in this code.

@author mstarch
"""
import fprime_gds.common.history.ram


class Histories:
    """
    Class to handle the individual histories. This handles the following histories:

    1. Channel history
    2. Event history
    3. Command history (short-circuited feedback from encoder)
    """

    def __init__(self):
        """ Constructor of histories composer """
        self._command_hist = None
        self._event_hist = None
        self._channel_hist = None

    def setup_histories(self, coders):
        """
        Setup a set of history objects in order to store the events of the decoders. This registers itself with the
        supplied coders object.

        :param coders: coders object to register histories with
        """
        # Create histories, RAM histories for now
        self._command_hist = fprime_gds.common.history.ram.RamHistory()
        self._event_hist = fprime_gds.common.history.ram.RamHistory()
        self._channel_hist = fprime_gds.common.history.ram.RamHistory()
        # Register histories where channels and packets are routed together
        coders.register_event_consumer(self._event_hist)
        coders.register_channel_consumer(self._channel_hist)
        coders.register_packet_consumer(self._channel_hist)
        coders.register_command_consumer(self._command_hist)

    @property
    def events(self):
        """
        Events history property
        """
        return self._event_hist

    @property
    def channels(self):
        """
        Channels history property
        """
        return self._channel_hist

    @property
    def commands(self):
        """
        Commands history property
        """
        return self._command_hist



####
#
####
import os
import flask_restful
import flask_restful.reqparse


class FlaskLogger(flask_restful.Resource):
    """
    Command dictionary endpoint. Will return dictionary when hit with a GET.
    """

    def __init__(self, logdir):
        """
        Constructor used to setup the log directory.

        :param logdir: log directory to search for logs
        """
        self.logdir = logdir

    def get(self):
        """
        Returns the logdir.
        """
        logs = {}
        listing = os.listdir(self.logdir)
        for path in [path for path in listing if path.endswith(".log")]:
            full_path = os.path.join(self.logdir, path)
            offset = 0
            with open(full_path) as file_handle:
                file_handle.seek(offset)
                logs[path] = file_handle.read()
        return logs



"""
flask/components.py:

This sets up the primary data components that allow Flask to connect into the system. This is where the standard
pipeline and other components are created to interact with Flask.
"""
import os

import fprime_gds.common.pipeline.standard

# Module variables, should remain hidden. These are singleton top-level objects used by Flask, and its various
# blueprints needed to run the system.
__PIPELINE = None


def setup_pipelined_components(
    debug, logger, config, dictionary, down_store, log_dir, tts_address, tts_port
):
    """
    Setup the standard pipeline and related components. This is done once, and then the resulting singletons are
    returned so that one object is used throughout the system.

    :param debug: used to prevent the construction of the standard pipeline
    :param logger: logger to use for output
    :param config: GDS configuration
    :param dictionary: path to F prime dictionary
    :param down_store:
    :param log_dir: log directory to write logs to, and serve logs from
    :param tts_address: address to the middleware layer
    :param tts_port: port of the middleware layer
    :return: F prime pipeline
    """
    global __PIPELINE
    if (
        __PIPELINE is None
        and not debug
        or os.environ.get("WERKZEUG_RUN_MAIN") == "true"
    ):
        pipeline = fprime_gds.common.pipeline.standard.StandardPipeline()
        pipeline.setup(config, dictionary, down_store, logging_prefix=log_dir)
        logger.info(
            "Connecting to GDS at: {}:{} from pid: {}".format(
                tts_address, tts_port, os.getpid()
            )
        )
        pipeline.connect(tts_address, tts_port)
        __PIPELINE = pipeline
    return __PIPELINE


def get_pipelined_components():
    """
    Returns the setup pipelined components, or raises exception if not setup yet.

    :return: F prime pipeline
    """
    assert __PIPELINE is not None, "Pipeline must be setup before use"
    return __PIPELINE



####
# channels.py:
#
# This file captures the HTML endpoint for the events list. This file contains one endpoint which allows for listing
# events after a certain time-stamp.
#
#  GET /events: list events
#      Input Data: {
#                      "start-time": "YYYY-MM-DDTHH:MM:SS.sss" #Start time for event listing
#                  }
####
import types

import flask_restful
import flask_restful.reqparse


class ChannelDictionary(flask_restful.Resource):
    """
    Channel dictionary endpoint. Will return dictionary when hit with a GET.
    """

    def __init__(self, dictionary):
        """
        Constructor used to setup for dictionary.
        """
        self.dictionary = dictionary

    def get(self):
        """
        Returns the dictionary object
        """
        return self.dictionary


class ChannelHistory(flask_restful.Resource):
    """
    Endpoint to return telemetry history data with optional time argument.
    """

    def __init__(self, history):
        """
        Constructor used to setup time argument to this history.

        :param history: history object holding channel
        """
        self.parser = flask_restful.reqparse.RequestParser()
        self.parser.add_argument(
            "session", required=True, help="Session key for fetching data."
        )
        self.history = history

    def get(self):
        """
        Return the telemetry history object
        """
        args = self.parser.parse_args()
        new_channels = self.history.retrieve(start=args.get("session"))
        self.history.clear()
        for chan in new_channels:
            # Add the 'display_text' to the event, along with a getter
            if chan.template.get_format_str() is not None:
                setattr(
                    chan,
                    "display_text",
                    chan.template.get_format_str() % (chan.val_obj.val),
                )

                def func(this):
                    return this.display_text
                setattr(chan, "get_display_text", types.MethodType(func, chan))
        return {"history": new_channels}

    def delete(self):
        """
        Delete the event history for a given session. This keeps the data all clear like.
        """
        args = self.parser.parse_args()
        self.history.clear(start=args.get("session"))



''' Versioning information for the GDS

This file contains versioning and compatibility definitions for the GDS tool set.
'''

# Currently the GDS is backwards compatible with all dictionaries. This MUST be bumped with each framework release to
# ensure compatibility
MINIMUM_SUPPORTED_FRAMEWORK_VERSION = (0,0,0)
MAXIMUM_SUPPORTED_FRAMEWORK_VERSION = (3,0,0)



"""
jsonable.py:

Folder with helper methods that allow for seamless conversion of F prime types to JSON types. This allows us to produce
data in JSON format with ease, assuming that the correct encoder is registered.

Note: JSON types must use only the following data types

 1. booleans
 2. numbers
 3. strings
 4. lists
 5. anonymous objects (dictionaries)

@author mstarch
"""


def fprime_to_jsonable(obj):
    """
    Takes an F prime object and converts it to a jsonable type.

    :param obj: object to convert
    :return: object in jsonable format (can call json.dump(obj))
    """
    # Otherwise try and scrape all "get_" getters in a smart way
    anonymous = {}
    getters = [attr for attr in dir(obj) if attr.startswith("get_")]
    for getter in getters:
        # Call the get_ functions, and call all non-static methods
        try:
            func = getattr(obj, getter)
            item = func()
            # If there is a property named "args" it needs to be handled specifically unless an incoming command
            if (
                getter == "get_args"
                and not "fprime_gds.common.data_types.cmd_data.CmdData"
                in str(type(obj))
            ):
                args = []
                for arg_spec in item:
                    arg_dict = {
                        "name": arg_spec[0],
                        "description": arg_spec[1],
                        "value": arg_spec[2].val,
                        "type": str(arg_spec[2]),
                    }
                    if arg_dict["type"] == "Enum":
                        arg_dict["possible"] = arg_spec[2].keys()
                    args.append(arg_dict)
                # Fill in our special handling
                item = args
            anonymous[getter.replace("get_", "")] = item
        except TypeError:
            continue
    return anonymous



"""
string_util.py

Utility functions to process strings to be used in FPrime GDS

@Created March 18, 2021
@janamian
"""

import re


def format_string(format_str, values):
    """
    Function to convert C-string style to python format
    without using python interpolation
    Considered the following format for C-string:
    %[flags][width][.precision][length]type
    
    %:           (?<!%)(?:%%)*%
    flags:       ([\-\+0\ \#])?
    width:       (\d+|\*)?
    .precision:  (\.\*|\.\d+)?
    length:      `([hLIw]|l{1,2}|I32|I64)?`
    type:        `([cCdiouxXeEfgGaAnpsSZ])`
    
    Note:
    This function will keep the flags, width, .precision and length of C-string
    template. It will remove all types so they could be duck-typed by python 
    interpreter except for hex type X or x.
    """
    def convert(match_obj):
        if match_obj.group() is not None:
            template = "{:" + match_obj.group()[1:-1]
            if match_obj.group()[-1].lower() == 'x':
                template += match_obj.group()[-1]
            template += "}"
            return template
        return match_obj

    pattern = "(?<!%)(?:%%)*%([\-\+0\ \#])?(\d+|\*)?(\.\*|\.\d+)?([hLIw]|l{1,2}|I32|I64)?([cCdiouxXeEfgGaAnpsSZ])"
    match = re.compile(pattern) 
    formated_str = re.sub(match, convert, format_str)
    result = formated_str.format(*values)
    return result



####
# json.py:
#
# Encodes GDS objects as JSON.
####
import collections
import enum

import flask.json


class GDSJsonEncoder(flask.json.JSONEncoder):
    """
    Custom class used to handle GDS object to JSON
    """

    def default(self, obj):
        """
        Override the default JSON encoder to pull out a dictionary for our handled types for encoding with the default
        encoder built into flask

        :param obj: obj to encode
        :return: JSON
        """
        # Object may already define a method for this, if so call it
        if hasattr(obj, "to_jsonable"):
            return obj.to_jsonable()
        # Dictionaries are "iterable", must handle them first
        elif isinstance(obj, collections.Mapping):
            return flask.json.JSONEncoder.default(self, obj)
        # Enumerations, just return the list of values
        elif isinstance(obj, enum.Enum):
            enum_dict = {"value": str(obj), "values": {}}
            for enum_val in type(obj):
                enum_dict["values"][str(enum_val)] = enum_val.value
            return enum_dict
        # Lists and iterables
        try:
            return list(iter(obj))
        except TypeError:
            pass
        return flask.json.JSONEncoder.default(self, obj)



####
# events.py:
#
# This file captures the HTML endpoint for the events list. This file contains one endpoint which allows for listing
# events after a certain time-stamp.
#
#  GET /events: list events
#      Input Data: {
#                      "start-time": "YYYY-MM-DDTHH:MM:SS.sss" #Start time for event listing
#                  }
####
import types

import flask_restful
import flask_restful.reqparse
from fprime_gds.common.utils.string_util import format_string


class EventDictionary(flask_restful.Resource):
    """
    Event dictionary endpoint. Will return dictionary when hit with a GET.
    """

    def __init__(self, dictionary):
        """
        Constructor used to setup for dictionary.
        """
        self.dictionary = dictionary

    def get(self):
        """
        Returns the dictionary object
        """
        return self.dictionary


class EventHistory(flask_restful.Resource):
    """
    Endpoint to return event history data with optional time argument.
    """

    def __init__(self, history):
        """
        Constructor used to setup time argument to this history.

        :param history: history object holding events
        """
        self.parser = flask_restful.reqparse.RequestParser()
        self.parser.add_argument(
            "session", required=True, help="Session key for fetching data."
        )
        self.history = history

    def get(self):
        """
        Return the event history object
        """
        args = self.parser.parse_args()
        new_events = self.history.retrieve(args.get("session"))
        self.history.clear()
        for event in new_events:
            # Add the 'display_text' to the event, along with a getter
            setattr(
                event,
                "display_text",
                format_string(event.template.format_str, tuple([arg.val for arg in event.args])),
            )

            def func(this):
                return this.display_text
            setattr(event, "get_display_text", types.MethodType(func, event))
        return {"history": new_events}

    def delete(self):
        """
        Delete the event history for a given session. This keeps the data all clear like.
        """
        args = self.parser.parse_args()
        self.history.clear(start=args.get("session"))



"""
__init__.py:

Python Logging Setup. This sets up the global logging format for all Gds components. This allows the user to use Python
loggers without calling basic config.

@author mstarch
"""
import logging
import os
import sys


def configure_py_log(directory=None, filename=sys.argv[0], mode="w"):
    """
    Configure the python logging. If logdir is supplied, our logs will go in that directory as a log file. Otherwise,
    logs will go to the CLI.

    :param filename: logging filename
    :param logdir: directory to log file into
    :param mode: of file to write
    """
    if directory is None:
        logging.basicConfig(
            level=logging.INFO,
            format="[%(asctime)s] [%(levelname)s] %(name)s: %(message)s",
        )
    else:
        logging.basicConfig(
            filename=os.path.join(directory, filename),
            filemode=mode,
            level=logging.INFO,
            format="[%(asctime)s] [%(levelname)s] %(name)s: %(message)s",
        )
